#pragma once
#include <vector>
#include <glad/glad.h>
#include <J3ML/LinearAlgebra/Vector2.hpp>
#include <J3ML/LinearAlgebra/Vector2i.hpp>
#include <J3ML/LinearAlgebra/Vector3.hpp>
#include <J3ML/LinearAlgebra/Vector4.hpp>
#include <Color4.hpp>

namespace JGL {
    class VRamList;
    /// A hint for OpenGL on how the VBO is to be used.
    enum VRamUsageHint : GLenum {
        // Never updated after creation.
        Fixed = GL_STATIC_DRAW,
        // Modified occasionally.
        Dynamic = GL_DYNAMIC_DRAW,
        // Constantly modified and used one or a few times.
        Stream = GL_STREAM_DRAW
    };
}

/// A wrapped for "Vertex Buffer Object" In OpenGL, Store things in V-ram.
class JGL::VRamList {
private:
    VRamUsageHint usage_hint = Fixed;
    GLuint list_handle = 0;
    long byte_count = 0;
    bool element_array_buffer = false;
    // TODO mutex lock.
    void load(const GLfloat* data, const long& size);
    void load(const GLuint* data, const long& size);
    void SetData(void* data, const long& count);
    void UpdateData(void* data, const long& offset, const long& count);
    void Erase();
public:
    VRamList(const GLuint* data, const long& count, VRamUsageHint hint = Fixed);
    VRamList(const GLfloat* data, const long& count, VRamUsageHint hint = Fixed);
    VRamList(const Vector2* data, const long& count, VRamUsageHint hint = Fixed);
    VRamList(const Vector3* data, const long& count, VRamUsageHint hint = Fixed);
    VRamList(const Vector4* data, const long& count, VRamUsageHint hint = Fixed);
    VRamList(const Color4* data, const long& count, VRamUsageHint hint = Fixed);
    /// Allocate an empty VBO
    /// @param byte_count the size of the buffer in bytes.
    /// @param element_array_buffer if applicable, whether the buffer is to be used for GLuint indices.
    /// @param hint A hint to the graphics driver for what the buffer is to be used for.
    VRamList(const size_t& byte_count, bool element_array_buffer, VRamUsageHint hint = Fixed);

    ~VRamList();
    /** Copying around the VBO data to a new VBO like this is slow.
     * Pass to function by const reference or pointer always. */
    VRamList(const VRamList& rhs);
    VRamList() : list_handle(0), byte_count(0), element_array_buffer(false) {}
public:
    [[nodiscard]] GLuint Handle() const;
    /// Returns the number of elements in the list.
    [[nodiscard]] long Length() const;
    /// Returns the size of the data in bytes.
    [[nodiscard]] size_t Size() const;
    /** Get VBO data back from the GPU. This is *bad* because the CPU is going to wait
    * for the transfer to finish. Has limited use other than testing. */
    [[nodiscard]] std::vector<GLfloat> GetDataF() const;
    [[nodiscard]] std::vector<GLuint> GetDataUI() const;
    /** Replace the data of an existing VBO in it's entirety. Must be same type. */
    void SetData(const GLfloat* data, const long& count);
    void SetData(const Vector2* data, const long& count);
    void SetData(const Vector3* data, const long& count);
    void SetData(const Vector4* data, const long& count);
    void SetData(const Color4* data, const long& count);

    void SetData(const GLuint* data, const long& count);
    void SetData(const Vector2i* data, const long& count);

    /** Update only a portion of the data in a VBO. Must be same type.
     * "offset" refers the number of Typename T into the buffer the data you want to change is.
     * For ex, offset 0 and length of 1 overwrites the first value. Offset 1 the second etc */

    // TODO provide a bool to specify whether the current buffer should be orphaned.
     void UpdateData(const GLfloat* data, const long& offset, const long& count);
     void UpdateData(const Vector2* data, const long& offset, const long& count);
     void UpdateData(const Vector3* data, const long& offset, const long& count);
     void UpdateData(const Vector4* data, const long& offset, const long& count);
     void UpdateData(const Color4* data, const long& offset, const long& count);
     void UpdateData(const GLuint* data, const long& offset, const long& count);
     void UpdateData(const Vector2i* data, const long& offset, const long& count);

     // Update only a portion of the data in a VBO using bytes.
     // TODO This version of the function has no protection for out of bounds writes.
     void UpdateData(const uint8_t* data, const long& offset, const long& count);
};