#include "internals.h"
#include <J3ML/LinearAlgebra/Vector4.hpp>
#include <JGL/types/Light.h>

// TODO handle the case that a required light is in the list of optional lights.
void JGL::J3D::SelectLights(const Vector3& position) {
    std::array<const LightBase*, 8> result = current_state.required_lights;

    unsigned int required_light_count = 0;
    for (const auto& i : result)
        if (i) required_light_count++;

    // If there is optional lights.
    if (!current_state.optional_lights.empty()) {
        // The number of lights we need to solve for
        unsigned int remaining_lights = 8 - required_light_count;

        std::vector<std::pair<float, const LightBase*>> light_influence;
        for (const auto* light: current_state.optional_lights)
            light_influence.emplace_back(light->GetAttenuationAtPosition(position), light);

        // Sort by biggest influence.
        std::sort(light_influence.begin(), light_influence.end(),
                  [](const auto &a, const auto &b) { return a.first > b.first; });

        // Add in optional lights.
        for (unsigned int i = 0; i < remaining_lights && i < light_influence.size(); i++)
            result[required_light_count++] = light_influence[i].second;
    }

    glEnable(GL_LIGHTING);
    for (unsigned int i = 0; i < result.size(); i++) {
        // Terminate early if we have less than 8 lights.
        if (!result[i])
            break;

        Vector4 ambient = { result[i]->GetAmbient().RN(), result[i]->GetAmbient().GN(), result[i]->GetAmbient().BN(), result[i]->GetAmbient().AN() };
        Vector4 diffuse = { result[i]->GetDiffuse().RN(), result[i]->GetDiffuse().GN(), result[i]->GetDiffuse().BN(), result[i]->GetDiffuse().AN() };
        Vector4 specular = { result[i]->GetSpecular().RN(), result[i]->GetSpecular().GN(), result[i]->GetSpecular().BN(), result[i]->GetSpecular().AN() };
        GLenum current_light = GL_LIGHT0 + i;

        glEnable(GL_LIGHT0 + i);
        glLightfv(current_light, GL_POSITION, result[i]->GetPosition().ptr());

        // How the light will affect different materials.
        glLightfv(current_light, GL_AMBIENT, ambient.ptr());
        glLightfv(current_light, GL_DIFFUSE, diffuse.ptr());
        glLightfv(current_light, GL_SPECULAR, specular.ptr());

        // How far the light goes.
        glLightf(current_light, GL_CONSTANT_ATTENUATION, result[i]->GetConstantAttenuation());
        glLightf(current_light, GL_LINEAR_ATTENUATION, result[i]->GetLinearAttenuation());
        glLightf(current_light, GL_QUADRATIC_ATTENUATION, result[i]->GetQuadraticAttenuation());
    }
}

void JGL::J3D::ResetLights() {
    Vector4 position = {0, 0, 1, 0};
    Vector4 ambient = {0, 0, 0, 1};
    Vector4 diffuse_light0 = {1,1,1,1};
    Vector4 diffuse = {0,0,0,1};
    Vector4 specular_light0 = {1,1,1,1};
    Vector4 specular = {0,0,0,1};
    Vector3 spot_direction = {0.0, 0.0, -1.0};

    for (unsigned int i = 0; i < 8; i++) {
        GLenum current_light = GL_LIGHT0 + i;

        glLightfv(current_light, GL_POSITION, position.ptr());
        glLightfv(current_light, GL_AMBIENT, ambient.ptr());
        glLightf(current_light, GL_CONSTANT_ATTENUATION, 1);
        glLightf(current_light, GL_LINEAR_ATTENUATION, 0);
        glLightf(current_light, GL_QUADRATIC_ATTENUATION, 0);
        glLightf(current_light, GL_SPOT_CUTOFF, 180.0);
        glLightfv(current_light, GL_SPOT_DIRECTION, spot_direction.ptr());
        glLightf(current_light, GL_SPOT_EXPONENT, 0.0);
        glLightfv(current_light, GL_DIFFUSE, (i == 0 ? diffuse_light0.ptr() : diffuse.ptr()));
        glLightfv(current_light, GL_SPECULAR, (i == 0 ? specular_light0.ptr() : specular.ptr()));
        glDisable(current_light);
    }
    glDisable(GL_LIGHTING);
}

bool JGL::J3D::UsingLighting() {
    if (!current_state.optional_lights.empty())
        return true;

    for (unsigned int i = 0; i < current_state.required_lights.size(); i++)
        if (current_state.required_lights[i])
            return true;
    return false;
}

std::vector<Vector3> JGL::TriangleMeshVertexNormals(const Vector3* vertices, const size_t& vertex_count, const unsigned int* indices, const size_t& index_count) {
    std::vector<Vector3> normals(vertex_count, Vector3(0, 0, 0));

    for (size_t i = 0; i < index_count; i += 3) {
        GLuint i1 = indices[i];
        GLuint i2 = indices[i + 1];
        GLuint i3 = indices[i + 2];

        Vector3 v1 = vertices[i1];
        Vector3 v2 = vertices[i2];
        Vector3 v3 = vertices[i3];

        Vector3 edge1 = v2 - v1;
        Vector3 edge2 = v3 - v1;

        Vector3 faceNormal = Vector3::Cross(edge1, edge2);

        normals[i1] += faceNormal;
        normals[i2] += faceNormal;
        normals[i3] += faceNormal;
    }
    for (auto& normal : normals)
        normal = normal.Normalized();

    return normals;
}

void JGL::StateStack::Push(const JGL::State& state) {
    states.push_back(state);
}

void JGL::StateStack::Pop() {
    states.pop_back();
}

JGL::State* JGL::StateStack::PreviousState() {
    if (states.empty())
        return nullptr;

    return &states.back();
}

JGL::State JGL::State::SaveState(const State& state) {
    State result;

    result.depth_test = glIsEnabled(GL_DEPTH_TEST);
    result.vertex_array = glIsEnabled(GL_VERTEX_ARRAY);
    result.normal_array = glIsEnabled(GL_NORMAL_ARRAY);
    result.cull_face = glIsEnabled(GL_CULL_FACE);
    result.blend = glIsEnabled(GL_BLEND);
    result.texture_2D = glIsEnabled(GL_TEXTURE_2D);
    result.texture_coordinate_array = glIsEnabled(GL_TEXTURE_COORD_ARRAY);
    result.color_array = glIsEnabled(GL_COLOR_ARRAY);

    glGetIntegerv(GL_ACTIVE_TEXTURE, &result.selected_texture_unit);
    result.selected_texture_unit -= GL_TEXTURE0;

    glGetFloatv(GL_COLOR_CLEAR_VALUE, result.clear_color);
    glGetFloatv(GL_CURRENT_COLOR, result.draw_color);
    glGetIntegerv(GL_VIEWPORT, result.viewport);
    glGetIntegerv(GL_FRAMEBUFFER_BINDING, (GLint*)&result.current_fbo);
    glGetIntegerv(GL_CURRENT_PROGRAM, &result.current_shader_handle);
    result.current_shader = state.current_shader;
    glGetIntegerv(GL_BLEND_SRC, &result.blend_func[0]);
    glGetIntegerv(GL_BLEND_DST, &result.blend_func[1]);

    return result;
}

void JGL::State::RestoreState(const State& state) {
    if (state.depth_test)
        glEnable(GL_DEPTH_TEST);
    else
        glDisable(GL_DEPTH_TEST);

    if (state.vertex_array)
        glEnableClientState(GL_VERTEX_ARRAY);
    else
        glDisableClientState(GL_VERTEX_ARRAY);

    if (state.normal_array)
        glEnableClientState(GL_NORMAL_ARRAY);
    else
        glDisableClientState(GL_NORMAL_ARRAY);

    if (state.cull_face)
        glEnable(GL_CULL_FACE);
    else
        glDisable(GL_CULL_FACE);

    if (state.blend)
        glEnable(GL_BLEND);
    else
        glDisable(GL_BLEND);

    glActiveTexture(GL_TEXTURE0 + state.selected_texture_unit);
    glClientActiveTexture(GL_TEXTURE0 + state.selected_texture_unit);

    if (state.texture_2D)
        glEnable(GL_TEXTURE_2D);
    else
        glDisable(GL_TEXTURE_2D);

    if (state.texture_coordinate_array)
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    else
        glDisableClientState(GL_TEXTURE_COORD_ARRAY);

    if (state.color_array)
        glEnableClientState(GL_COLOR_ARRAY);
    else
        glDisableClientState(GL_COLOR_ARRAY);

    if (state.current_shader)
        glUseProgram(state.current_shader->Handle());
    else
        glUseProgram(0);

    glBindFramebuffer(GL_FRAMEBUFFER, state.current_fbo);
    glViewport(state.viewport[0], state.viewport[1], state.viewport[2], state.viewport[3]);
    glClearColor(state.clear_color[0], state.clear_color[1], state.clear_color[2], state.clear_color[3]);
    glColor4f(state.draw_color[0], state.draw_color[1], state.draw_color[2], state.draw_color[3]);
    glBlendFunc(state.blend_func[0], state.blend_func[1]);
}
