
#include <vector>
#include <string>
#include <iostream>
#include <glad/glad.h>
#include <JGL/logger/logger.h>
#include <JGL/JGL.h>

#if __linux__
    #include <freetype2/ft2build.h>
    #include FT_FREETYPE_H
    #include FT_OUTLINE_H
#endif

#if _WIN32
    #include <ft2build.h>
    #include FT_FREETYPE_H
    #include FT_OUTLINE_H
#endif

#include <JGL/types/Font.h>
#include <JGL/types/FontCache.h>

namespace JGL::Detail
{
    FT_Library ft;

    std::vector<Font> fonts;

    bool InitTextEngine() {

        glPixelStorei(GL_UNPACK_ALIGNMENT, 1); // NOTE: This MUST be called for text rendering to work properly!!!
        // Keep note of this, might cause problems later?

        if (ft != nullptr)
            Logger::Fatal("Error::FREETYPE: FT_Library was initialized but is already initialized.");
        
        if (FT_Init_FreeType(&ft))
            return true;
        return false;
    }

    void UnloadFont(int font_index) {

        auto iter = fonts.begin();
        while (iter != fonts.end())
        {
            if (iter->index == font_index){
                FT_Done_Face(iter->face);
                iter = fonts.erase(iter);
            } else ++iter;
        }
    }
}

namespace JGL {

    bool InitTextEngine() {
        return Detail::InitTextEngine();
    }

    Font::Font(const unsigned char* data, const size_t& size) {
        if (Detail::ft == nullptr)
            throw std::runtime_error("Error::FREETYPE: FT_Library was not initialized before attempting to load a font!");

        if (FT_New_Memory_Face(Detail::ft, data, size, 0, &face))
            throw std::runtime_error("Error::FREETYPE: Failed to load font!");

        unsigned int new_index = 0;
        for (const auto& f : Detail::fonts)
            if (f.index >= new_index)
                new_index = f.index + 1;
        index = new_index;

        Detail::fonts.push_back(*this);
        std::cout << "Loaded font from memory at " << static_cast<const void*>(data) << " with index " << new_index << std::endl;
    }

    Font::Font(const std::filesystem::path& path) {
        if (Detail::ft == nullptr)
            throw std::runtime_error("Error::FREETYPE: FT_Library was not initialized before attempting to load a font!");

        if (FT_New_Face(Detail::ft, path.string().c_str(), 0, &face))
            throw std::runtime_error("Error::FREETYPE: Failed to load font!");

        unsigned int new_index = 0;
        for (const auto& f : Detail::fonts)
            if (f.index >= new_index)
                new_index = f.index + 1;

        index = new_index;
        Detail::fonts.push_back(*this);
        std::cout << "Loaded font from " << path << " with index " << new_index << std::endl;
    }

    Font::~Font() {
        //Detail::UnloadFont(this->index);
    }


    std::vector<Font> Font::GetLoadedFonts() {
        return Detail::fonts;
    }


    Font Font::LoadTTF(const std::filesystem::path& path)
    {
        return Font(path);
    }

    Vector2 Font::MeasureString(const std::string& text, unsigned int ptSize) {
        Vector2 extents = Vector2::Zero;

        for(auto& f : fontCache.getFonts()) {
            if (f->getFontIndex() == this->index) {
                for (const char &c: text) {
                    auto glyph = f->getGlyph(c);
                    extents.x += glyph->advanceX;
                    extents.y = glyph->ascent + glyph->descent;
                }
                return extents;
            }
        }

        // No cache
        FT_Set_Pixel_Sizes(this->face, ptSize, ptSize);
        for (const char& c : text) {
            FT_GlyphSlot slot = face->glyph;
            auto glyph_index = FT_Get_Char_Index(this->face, c);
            auto error = FT_Load_Glyph(this->face, glyph_index, FT_LOAD_DEFAULT);

            if (error)
                continue;

            extents.x += static_cast<float>(slot->advance.x >> 6);
            extents.y = (face->size->metrics.ascender / 64.0f) + (-face->size->metrics.descender / 64.0f);
        }
        return extents;
    }
}