#include <JGL/types/Shader.h>

#include <glad/glad.h>

namespace JGL {

    class ShaderPreprocessor {
    public:
        std::string include_keyword = "#include";
        ShaderPreprocessor() {}
        std::string LoadShaderFile(const std::filesystem::path& filepath)
        {
            // if size is 0, it indicates, that we are at the top of the recursive load stack
            bool stack_top = (already_included.size() == 0);

            if (stack_top) {
                already_included.emplace_back(filepath.string());
            }

            std::string ret_data = "";
            std::ifstream file(filepath);

            if (!file.good())
            {
                std::cout << "ERROR [ ShaderLoader::load_shader ]: Failed to start fstream, check if '" << filepath << "' exists\n";
                return ret_data;
            }

            if (file.is_open()) {
                std::string line;
                while (getline(file, line)) {
                    if (line.find(include_keyword) == 0)
                    {
                        // Get path between double quotes
                        std::string rel_include_path = extract_first_between(line.substr(include_keyword.length()), '"', '"');

                        // Modify path according to os
                        // TODO: Use std::filesystem dummy...
                        // later, let me get this stolen code to work first.
                    #ifdef _WIN32
                        std::replace(rel_include_path.begin(), rel_include_path.end(), '/', '\\');
                    #elif __linux__
                        std::replace(rel_include_path.begin(), rel_include_path.end(), '\\', '/');
                    #endif
                        std::string full_include_path = extract_path(filepath.string()) + rel_include_path;

                        // Avoid including self
                        if (filepath == full_include_path) {
                            std::cout << "WARNING [ ShaderLoader::load_shader ]: '"<< filepath <<"' tried to include itself\n";
                            continue;
                        } else {
                            bool include_flag = true;
                            // check if the current file is already included
                            for (const auto& file_to_check : already_included) {
                                // Avoid duplicate includes
                                if (file_to_check == full_include_path) {
                                    include_flag = false;
                                    break;
                                }
                            }

                            // If not yet included, push path to include vector and replace line with file contents
                            if (include_flag) {
                                already_included.push_back(full_include_path);
                                // Repeat recursively.
                                ret_data += LoadShaderFile(full_include_path) + '\n';
                            }
                        }
                    } else {
                        ret_data += line + "\n";
                    }
                }
                file.close();
            } else {
                std::cout << "ERROR [ ShaderLoader::load_shader ]: Unable to open file '" << filepath << "'\n";
            }

            // We are back to the first call
            if (stack_top) {
                already_included.clear();
            }
            return ret_data;
        }




        std::string PreprocessShaderSource(const std::string& source);
    private:
        std::vector<std::string> already_included;
        /// Helper function that strips filename from a path, basically getting the path to the directory containing the file.
        std::string extract_path(const std::string& path)
        {
            // find the position of the last directory separator.
            std::size_t pos = path.find_last_of("\\/");

            // strip fname from path
            if (pos != std::string::npos) {
                return path.substr(0, pos+1);
            }
            return "";
        }
        std::string extract_first_between(const std::string& input, char start_symbol, char end_symbol)
        {
            size_t start_index = input.find(start_symbol);
            size_t end_index = input.find(end_symbol, start_index+1);

            std::string extracted = "";
            if (start_index != std::string::npos && end_index != std::string::npos) {
                extracted = input.substr(start_index + 1, end_index - start_index - 1);
            } else {
                std::cout << "ERROR [ ShaderLoader::extract_first_between ]: Start '" << start_symbol << "' or end symbol '" << end_symbol << "' not found" << std::endl;
            }
            return extracted;
        }
    };

    Shader::Shader(const std::filesystem::path& vertex_source_path, const std::filesystem::path& fragment_source_path, const std::vector<std::pair<std::string, GLint>>& attribute_bindings) :
    Shader(ReadFile(vertex_source_path), ReadFile(fragment_source_path), attribute_bindings) {
        vertexPath = vertex_source_path.string();
        fragmentPath = fragment_source_path.string();
    }

    void Shader::checkCompileErrors(GLuint shader, const std::string& type) {
        GLint success;
        GLchar infoLog[1024];

        if (type != "PROGRAM") {
            glGetShaderiv(shader, GL_COMPILE_STATUS, &success);
            if (!success) {
                glGetShaderInfoLog(shader, 1024, nullptr, infoLog);
                OnCompilationErrorMessage.Invoke( std::format("COMPILATION_ERROR: {}", type), infoLog);
            }
        } else {
            glGetProgramiv(shader, GL_LINK_STATUS, &success);
            if (!success) {
                glGetProgramInfoLog(shader, 1024, nullptr, infoLog);
                OnCompilationErrorMessage.Invoke(std::format("COMPILATION_ERROR: {}", type), infoLog);
            }
        }

        if (success)
            std::cout << "Shader compiled successfully" << std::endl;
    }

    GLint Shader::Uniform(const std::string &name) const {
        auto it = uniform_location_cache.find(name);
        if (it == uniform_location_cache.end()) {
            auto location = glGetUniformLocation(id, name.c_str());
            uniform_location_cache[name] = location;
            return location;
        }
        return it->second;
    }

    GLint Shader::Attribute(const std::string &name) const {
        return glGetAttribLocation(id, name.c_str());
    }

    Shader::Shader(const std::string &vertex_code, const std::string &fragment_code, const std::vector<std::pair<std::string, GLint>>& attribute_bindings) {
        vertexSource = vertex_code;
        fragmentSource = fragment_code;

        const char* vShaderCode = vertex_code.c_str();
        const char* fShaderCode = fragment_code.c_str();

        // 2. Compile shaders.
        unsigned int vertex, fragment;

        // vertex shader.
        vertex = glCreateShader(GL_VERTEX_SHADER);
        glShaderSource(vertex, 1, &vShaderCode, nullptr);
        glCompileShader(vertex);
        checkCompileErrors(vertex, "VERTEX");

        // fragment shader
        fragment = glCreateShader(GL_FRAGMENT_SHADER);
        glShaderSource(fragment, 1, &fShaderCode, nullptr);
        glCompileShader(fragment);
        checkCompileErrors(fragment, "FRAGMENT");

        // shader Program
        id = glCreateProgram();

        // bind attribute locations.
        for (auto& a: attribute_bindings)
            glBindAttribLocation(id, a.second, a.first.c_str());

        glAttachShader(id, vertex);
        glAttachShader(id, fragment);
        glLinkProgram(id);
        checkCompileErrors(id, "PROGRAM");

        // delete the shaders as they're linked into our program now and are no longer necessary
        glDeleteShader(vertex);
        glDeleteShader(fragment);
    }

    bool Shader::Loaded() const { return id != 0; }

    unsigned int Shader::Handle() const { return id;}

    void Shader::SetBool(const std::string &name, bool value) const { glUniform1i(Uniform(name), (int)value); }

    void Shader::SetInt(const std::string &name, int value) const { glUniform1i(Uniform(name), value); }

    void Shader::SetFloat(const std::string &name, float value) const { glUniform1f(Uniform(name), value); }

    void Shader::SetVector2(const std::string &name, const Vector2 &value) const { glUniform2f(Uniform(name), value.x, value.y); }

    void Shader::SetVector2(const std::string &name, float x, float y) const { glUniform2f(Uniform(name), x, y); }

    void Shader::SetVector3(const std::string &name, const Vector3 &value) const { glUniform3f(Uniform(name), value.x, value.y, value.z); }

    void Shader::SetVector3(const std::string &name, float x, float y, float z) const { glUniform3f(Uniform(name), x, y, z); }

    void Shader::SetVector4(const std::string &name, const Vector4 &value) const { glUniform4f(Uniform(name), value.x, value.y, value.z, value.w); }

    void Shader::SetVector4(const std::string &name, float x, float y, float z, float w) const { glUniform4f(Uniform(name), x, y, z, w); }

    void Shader::SetMatrix2x2(const std::string &name, const Matrix2x2 &value) const {
        /// TODO: Verify if glsl expects row-major or col-major!!
        bool transpose = false;
        glUniformMatrix2fv(Uniform(name), 4, transpose, value.ptr());
    }

    void Shader::SetMatrix3x3(const std::string &name, const Matrix3x3 &value) const {
        /// TODO: Verify if glsl expects row-major or col-major!!
        bool transpose = false;
        glUniformMatrix3fv(Uniform(name), 9, transpose, value.ptr());
    }

    void Shader::SetMatrix4x4(const std::string &name, const Matrix4x4 &value) const {
        /// TODO: Verify if glsl expects row-major or col-major!!
        bool transpose = false;
        glUniformMatrix4fv(Uniform(name), 16, transpose, value.ptr());
    }

    std::string Shader::ReadFile(const std::filesystem::path &path) {
        /*std::string contents;

        std::ifstream file;

        file.exceptions(std::ifstream::failbit | std::ifstream::badbit);

        try {
            // Open Files
            file.open(path);
            std::stringstream stream;

            // Read file's buffer contents into streams
            stream << file.rdbuf();

            // close file handlers
            file.close();

            // convert stream into string
            contents = stream.str();

            return contents;
        } catch (std::ifstream::failure& e) {
            std::cout << "ERROR::FILE_READ_FAILURE: " << e.what() << std::endl;
            return "";
        }*/

        ShaderPreprocessor processor;
        return processor.LoadShaderFile(path);
    }
}