#include <JGL/types/Texture.h>
#include <JGL/types/RenderTarget.h>
#include <JGL/logger/logger.h>
#include <glad/glad.h>
#include <fstream>

#define STB_IMAGE_IMPLEMENTATION
#include "internals/stb_image.h"

using namespace JGL;

Texture::Texture(const std::filesystem::path& file, FilteringMode filtering_mode, SampleRate anisotropy, WrappingMode wrapping_mode, bool invert_y) :
invert_y(invert_y), filtering_mode(filtering_mode), anisotropy(anisotropy), wrapping_mode(wrapping_mode) {
    std::vector<unsigned char> pixels{};

    std::ifstream ifStream(file, std::ios::in | std::ios::binary);
    if (!ifStream.is_open())
        Logger::Fatal("Trying to load a texture from: " + file.string() + "but we couldn't read the file.");

    unsigned char bmpFileHeader[14];

    ifStream.read(reinterpret_cast<char*>(bmpFileHeader), 14);
    if (bmpFileHeader[0] == 'B' && bmpFileHeader[1] == 'M')
        pixels = bmp(file);

    //TODO determine if it's a PNG using the file header instead.
    else if (file.string().ends_with(".png"))
        pixels = png(file);
    ifStream.close();


    if (invert_y) {
        unsigned int row_size = size.x * 4;
        if (format == ColorFormat::RGB)
            row_size = size.x * 3;

        std::vector<unsigned char> temp(row_size);
        for (unsigned int y = 0; y < size.y / 2; ++y) {
            unsigned char *topRow = &pixels[y * row_size];
            unsigned char *bottomRow = &pixels[(size.y - y - 1) * row_size];
            std::copy(bottomRow, bottomRow + row_size, temp.begin());
            std::copy(topRow, topRow + row_size, bottomRow);
            std::copy(temp.begin(), temp.end(), topRow);
        }
    }

    load(pixels.data());
}

std::vector<unsigned char> Texture::png(const std::filesystem::path& file) {
    std::vector<unsigned char> result{};
    int channels;

    unsigned char* image_data = stbi_load(file.string().c_str(), &size.x, &size.y, &channels, 0);
    if (!image_data)
        Logger::Fatal("Trying to load a texture from: " + file.string() + "but we couldn't read the file.");

    if (channels == 4)
        format = ColorFormat::RGBA;
    else if (channels == 3)
        format = ColorFormat::RGB;

    if (image_data)
        result.assign(image_data, image_data + size.x * size.y * channels);

    stbi_image_free(image_data);
    return result;
}

std::vector<unsigned char> Texture::bmp(const std::filesystem::path& file) {
    std::vector<unsigned char> result{};
    std::ifstream bmp_file(file, std::ios::in | std::ios::binary);

    unsigned char bmpFileHeader[14];
    unsigned char bmpInfoHeader[40];

    bmp_file.read(reinterpret_cast<char *>(bmpFileHeader), 14);
    bmp_file.read(reinterpret_cast<char *>(bmpInfoHeader), 40);

    size.x = bmpInfoHeader[4] + (bmpInfoHeader[5] << 8) + (bmpInfoHeader[6] << 16) + (bmpInfoHeader[7] << 24);
    size.y = bmpInfoHeader[8] + (bmpInfoHeader[9] << 8) + (bmpInfoHeader[10] << 16) + (bmpInfoHeader[11] << 24);

    int row_padded = (size.x * 3 + 3) & (~3);
    result.resize(size.x * size.y * 3);

    std::vector<unsigned char> rowData(row_padded);
    for (int y = size.y - 1; y >= 0; --y) {
        bmp_file.read(reinterpret_cast<char *>(rowData.data()), row_padded);
        for (int x = 0; x < size.x; ++x) {
            result[(y * size.x + x) * 3 + 2] = rowData[x * 3 + 0];
            result[(y * size.x + x) * 3 + 1] = rowData[x * 3 + 1];
            result[(y * size.x + x) * 3 + 0] = rowData[x * 3 + 2];
        }
    }

    bmp_file.close();
    format = ColorFormat::RGB;
    return result;
}

Texture::Texture(const Vector2i& size, FilteringMode filtering_mode) : invert_y(true), format(ColorFormat::RGBA), size(size), filtering_mode(filtering_mode), wrapping_mode(WrappingMode::CLAMP_TO_EDGE) {
    if (SizeExceedsMaximum(size))
        Logger::Error("Creating a texture where the size is bigger than the maximum for this system, use Texture::MaximumSize() to check for this beforehand.");
    GLuint previous_texture;
    glGetIntegerv(GL_TEXTURE_BINDING_2D, (GLint*) &previous_texture);

    glGenTextures(1, &texture_handle);
    glBindTexture(GL_TEXTURE_2D, texture_handle);
    //NEAREST
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

    if (filtering_mode == FilteringMode::NEAREST)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST),
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

    else if (filtering_mode == FilteringMode::BILINEAR)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR),
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

    if (filtering_mode == FilteringMode::MIPMAP_NEAREST)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST_MIPMAP_NEAREST),
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

    else if (filtering_mode == FilteringMode::MIPMAP_BILINEAR)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST),
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

    else if (filtering_mode == FilteringMode::MIPMAP_TRILINEAR)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR),
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

    //Clamp
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, size.x, size.y, 0, GL_RGBA, GL_UNSIGNED_BYTE, nullptr);
    glBindTexture(GL_TEXTURE_2D, previous_texture);
}
void Texture::load(const unsigned char* pixels) {
    // TODO a static member function to query for this.
    if (SizeExceedsMaximum(size))
        Logger::Error("Creating a texture where the size is bigger than the maximum for this system.");

    GLuint previous_texture;
    glGetIntegerv(GL_TEXTURE_BINDING_2D, (GLint*) &previous_texture);

    glGenTextures(1, &texture_handle);
    glBindTexture(GL_TEXTURE_2D, texture_handle);

    if (format == ColorFormat::RGBA)
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, size.x, size.y, 0, GL_RGBA, GL_UNSIGNED_BYTE, pixels);

    else if (format == ColorFormat::RGB)
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, size.x, size.y, 0, GL_RGB, GL_UNSIGNED_BYTE, pixels);

    if (wrapping_mode == WrappingMode::CLAMP_TO_EDGE)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE),
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

    else if (wrapping_mode == WrappingMode::REPEAT)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT),
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

    else if (wrapping_mode == WrappingMode::MIRRORED_REPEAT)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_MIRRORED_REPEAT),
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_MIRRORED_REPEAT);

    else if (wrapping_mode == WrappingMode::CLAMP_TO_BORDER)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER),
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER);

    if (filtering_mode == FilteringMode::MIPMAP_NEAREST ||
    filtering_mode == FilteringMode::MIPMAP_BILINEAR ||
    filtering_mode == FilteringMode::MIPMAP_TRILINEAR) {

        glGenerateMipmap(GL_TEXTURE_2D);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_BASE_LEVEL, 0);

        if (filtering_mode == FilteringMode::MIPMAP_NEAREST)
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST_MIPMAP_NEAREST),
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

        else if (filtering_mode == FilteringMode::MIPMAP_BILINEAR)
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST),
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

        else if (filtering_mode == FilteringMode::MIPMAP_TRILINEAR)
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR),
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

        if (anisotropy != SampleRate::NONE) {
            if (anisotropy > MaxAnisotropySampleRate())
                Logger::Error("Anisotropy set higher than the maximum value for this system, disabled, use Texture::MaxAnisotropy()."),
                anisotropy = SampleRate::NONE;
            else
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY, (int) anisotropy);
        }
    }

    else if (filtering_mode == FilteringMode::NEAREST || filtering_mode == FilteringMode::BILINEAR) {
        if (filtering_mode == FilteringMode::NEAREST)
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST),
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

        else if (filtering_mode == FilteringMode::BILINEAR)
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR),
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

        if (anisotropy != SampleRate::NONE)
            Logger::Error("Anisotropy only applies when using mipmaps with a 3D perspective, disabled."),
            anisotropy = SampleRate::NONE;
    }
    glBindTexture(GL_TEXTURE_2D, previous_texture);
}

std::vector<Color4> JGL::Texture::GetPixelData() const {
    GLint current_texture;
    glGetIntegerv(GL_TEXTURE_BINDING_2D, &current_texture);

    std::vector<Color4> result((size_t) (size.x * size.y));
    glBindTexture(GL_TEXTURE_2D, texture_handle);

    if (format == ColorFormat::RGBA) {
        glGetTexImage(GL_TEXTURE_2D, 0, GL_RGBA, GL_UNSIGNED_BYTE, result.data());
        glBindTexture(GL_TEXTURE_2D, current_texture);
        return result;
    }

    //if RGB
    std::vector<Color3> color3((size_t) (size.x * size.y));
    glGetTexImage(GL_TEXTURE_2D, 0, GL_RGB, GL_UNSIGNED_BYTE, color3.data());
    for (const auto &c: color3)
        result.emplace_back(c);

    glBindTexture(GL_TEXTURE_2D, current_texture);
    return result;
}

GLuint Texture::GetHandle() const {
    return texture_handle;
}

Vector2i Texture::GetDimensions() const {
    return size;
}

bool Texture::Inverted() const {
    return invert_y;
}

ColorFormat Texture::GetFormat() const {
    return format;
}

FilteringMode Texture::GetFilteringMode() const {
    return filtering_mode;
}

WrappingMode Texture::GetWrappingMode() const {
    return wrapping_mode;
}

Texture::Texture(const Color4* pixels, const Vector2i& size, FilteringMode filtering_mode, SampleRate anisotropy, WrappingMode wrapping_mode) :
 size(size), format(ColorFormat::RGBA), filtering_mode(filtering_mode), anisotropy(anisotropy), wrapping_mode(wrapping_mode) {
    load((unsigned char*) pixels);
}

Texture::Texture(const Color3* pixels, const Vector2i& size, FilteringMode filtering_mode, SampleRate anisotropy, WrappingMode wrapping_mode) :
size(size), format(ColorFormat::RGB), filtering_mode(filtering_mode), anisotropy(anisotropy), wrapping_mode(wrapping_mode) {
    load((unsigned char*) pixels);
}

Texture::~Texture() {
    if (texture_handle != 0)
        glDeleteTextures(1, &texture_handle);
    texture_handle = 0;
}

Texture::Texture(const Texture& rhs) {
    auto* this_texture = new Texture(rhs.GetDimensions());
    auto this_render_target = RenderTarget(this_texture);
    auto rhs_render_target = RenderTarget(&rhs);

    RenderTarget::Blit(rhs_render_target, &this_render_target);

    this->texture_handle = this_texture->texture_handle;
    this->size = this_texture->size;
    this->invert_y = this_texture->invert_y;
    this->format = this_texture->format;
    this->filtering_mode = this_texture->filtering_mode;
    this->wrapping_mode = this_texture->wrapping_mode;
    operator delete(this_texture);
}

bool Texture::SizeExceedsMaximum(const Vector2i& s) {

    auto max_size = Texture::MaxSize();
    return s.x > max_size.x || s.y > max_size.y;
}

Vector2i Texture::MaxSize() {
    GLint max_size;
    glGetIntegerv(GL_MAX_TEXTURE_SIZE, &max_size);

    return { max_size, max_size };
}

SampleRate Texture::MaxAnisotropySampleRate() {
    if (!GLAD_GL_ARB_texture_filter_anisotropic)
        return SampleRate::NONE;

    float anisotropy;
    glGetFloatv(GL_MAX_TEXTURE_MAX_ANISOTROPY, &anisotropy);

    if (anisotropy == 2)
        return SampleRate::X2;
    else if (anisotropy == 4)
        return SampleRate::X4;
    else if (anisotropy == 8)
        return SampleRate::X8;
    else
        return SampleRate::X16;
}

SampleRate Texture::GetAnisotropySampleRate() const {
    return anisotropy;
}
