#include <JGL/types/TextureAtlas.h>
#include <JGL/types/RenderTarget.h>
#include <JGL/logger/logger.h>

using namespace JGL;
AtlasRegion TextureAtlas::GetRegion(unsigned int region) const {
    if (region >= regions.size()) {
        Logger::Error("Requesting region " + std::to_string(region) + "but it's outside of the texture atlas?");
        return { {0, 0},{0, 0} };
    }
    return regions[region];
}

TextureAtlas::TextureAtlas(const Color4** pixels, const Vector2i** sizes, unsigned int texture_count, FilteringMode filtering_mode) {
    std::vector<Texture> textures(texture_count);

    int x_length = 0; int largest_height = 0;
    for (unsigned int i = 0; i < texture_count; i++) {
        textures[i] = Texture(pixels[i], *sizes[i], FilteringMode::NEAREST);
        x_length += sizes[i]->x;

        if (sizes[i]->y > largest_height)
            largest_height = sizes[i]->y;
    }

    int rows = 1;
    while (x_length > Texture::MaxSize().x)
        x_length /= 2, rows += 1;

    auto* result = new Texture(Vector2i(x_length, largest_height * rows), filtering_mode);
    auto* render_target = new RenderTarget(result);

    int current_x = 0; int current_y = 0;
    for (unsigned int i = 0; i < texture_count; i++) {
        if (current_x + sizes[i]->x > x_length)
            current_x = 0, current_y += largest_height;

        RenderTarget::Blit(&textures[i], render_target, Vector2i(current_x, current_y));
        regions.emplace_back(AtlasRegion({ current_x, current_y }, { sizes[i]->x, sizes[i]->y }));
        current_x += sizes[i]->x;
    }

    render_target->RegenerateMipMaps();
    delete render_target;

    this->texture_handle = result->GetHandle();
    this->size = result->GetDimensions();
    this->invert_y = result->Inverted();
    this->filtering_mode = result->GetFilteringMode();
    this->wrapping_mode = result->GetWrappingMode();
    this->anisotropy = result->GetAnisotropySampleRate();
    this->format = result->GetFormat();

    // Don't call destructor so the v-ram isn't cleared.
    operator delete(result);
}

TextureAtlas::TextureAtlas(const Color4* pixels, const Vector2i& size, AtlasRegion** regions, unsigned int region_count, FilteringMode filtering_mode, SampleRate anisotropy) : Texture(pixels, size, filtering_mode, anisotropy) {
    this->regions.resize(region_count);
    for (unsigned int i = 0; i < region_count; ++i)
        this->regions[i] = *regions[i];
}

