#include <JGL/types/VRamList.h>
#include <JGL/logger/logger.h>
#include <cstring>

// TODO combine the two load functions.

void JGL::VRamList::load(const GLfloat* data, const long& size) {

    GLint current_array_buffer = 0;
    glGetIntegerv(GL_ARRAY_BUFFER_BINDING, &current_array_buffer);
    glGenBuffers(1, &list_handle);
    glBindBuffer(GL_ARRAY_BUFFER, list_handle);
    glBufferData(GL_ARRAY_BUFFER, size, data, usage_hint);
    glBindBuffer(GL_ARRAY_BUFFER, current_array_buffer);
    element_array_buffer = false;
    byte_count = size;

}


void JGL::VRamList::load(const GLuint* data, const long& size) {

    GLint current_element_array_buffer = 0;
    glGetIntegerv(GL_ELEMENT_ARRAY_BUFFER_BINDING, &current_element_array_buffer);
    glGenBuffers(1, &list_handle);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, list_handle);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, size, data, usage_hint);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, current_element_array_buffer);
    element_array_buffer = true;
    byte_count = size;

}

void JGL::VRamList::Erase() {
    if (list_handle == 0)
        return;


    GLint current_element_array_buffer = 0;
    glGetIntegerv(GL_ELEMENT_ARRAY_BUFFER_BINDING, &current_element_array_buffer);
    GLint current_array_buffer = 0;
    glGetIntegerv(GL_ARRAY_BUFFER_BINDING, &current_array_buffer);

    if (element_array_buffer && current_element_array_buffer == list_handle)
        JGL::Logger::Warning("Erasing an element array buffer while it's in use?");

    else if (!element_array_buffer && current_array_buffer == list_handle)
        JGL::Logger::Warning("Erasing an array buffer while it's in use?");

    glDeleteBuffers(1, &list_handle);
    list_handle = 0;

}

GLuint JGL::VRamList::Handle() const {
    return list_handle;
}

long JGL::VRamList::Length() const {
    // sizeof(GLfloat) & sizeof(GLuint) are both 4 - Redacted.
    return byte_count / 4;
}

size_t JGL::VRamList::Size() const {
    return byte_count;
}

void JGL::VRamList::SetData(void* data, const long& length) {

    bool should_resize = (this->byte_count != length * 4);
    if (should_resize) {
        glDeleteBuffers(1, &list_handle);
        list_handle = 0;
        element_array_buffer ? load((GLuint*) data, sizeof(GLuint) * length) : load((GLfloat*) data, sizeof(GLfloat) * length);
        return;
    }

    // if (!should resize)
    GLint current_buffer = 0;
    GLenum buffer_type = GL_ARRAY_BUFFER;
    GLenum buffer_binding = GL_ARRAY_BUFFER_BINDING;
    if (element_array_buffer)
        buffer_type = GL_ELEMENT_ARRAY_BUFFER,
        buffer_binding = GL_ELEMENT_ARRAY_BUFFER_BINDING;

    glGetIntegerv(buffer_binding, &current_buffer);
    glBindBuffer(buffer_type, list_handle);

    size_t element_size = element_array_buffer ? sizeof(GLuint) : sizeof(GLfloat);

    glBufferSubData(buffer_type, 0, length * element_size, data);
    glBindBuffer(buffer_type, current_buffer);
}

void JGL::VRamList::UpdateData(void* data, const long& offset, const long& length) {

    if (offset + length > Length()) {
        unsigned long oob_delta = (offset + length) - Length();

        if (element_array_buffer) {
            auto list_data = GetDataUI();
            list_data.resize(list_data.size() + oob_delta);
            memcpy(list_data.data() + (offset * sizeof(GLuint)), data, length * sizeof(GLuint));
            // This is going unlock and relock really fast, But this code fixes something considered wrong anyway - Redacted.
            return SetData(list_data.data(), list_data.size());
        }

        // if (!element_array_buffer)
        auto list_data = GetDataF();
        list_data.resize(list_data.size() + oob_delta);
        memcpy(list_data.data() + (offset * sizeof(GLfloat)), data, length * sizeof(GLfloat));
        return SetData(list_data.data(), list_data.size());
    }

    GLint current_buffer = 0;
    GLenum buffer_type = GL_ARRAY_BUFFER;
    GLenum buffer_binding = GL_ARRAY_BUFFER_BINDING;

    if (element_array_buffer)
        buffer_type = GL_ELEMENT_ARRAY_BUFFER,
        buffer_binding = GL_ELEMENT_ARRAY_BUFFER_BINDING;

    glGetIntegerv(buffer_binding, &current_buffer);
    glBindBuffer(buffer_type, list_handle);

    size_t element_size = element_array_buffer ? sizeof(GLuint) : sizeof(GLfloat);

    glBufferSubData(buffer_type, offset * element_size, length * element_size, data);
    glBindBuffer(buffer_type, current_buffer);
}

std::vector<GLfloat> JGL::VRamList::GetDataF() const {

    if (element_array_buffer)
        JGL::Logger::Warning("Getting data as GLfloat but the buffer data is GLuint?");

    bool vertex_array_enabled = glIsEnabled(GL_VERTEX_ARRAY);
    if (!vertex_array_enabled)
        glEnableClientState(GL_VERTEX_ARRAY);

    GLint current_buffer = 0;
    glGetIntegerv(GL_ARRAY_BUFFER_BINDING, &current_buffer);
    glBindBuffer(GL_ARRAY_BUFFER, list_handle);

    std::vector<GLfloat> data(byte_count / 4);
    void* vram = glMapBuffer(GL_ARRAY_BUFFER, GL_READ_ONLY);
    if (vram == nullptr)
        JGL::Logger::Fatal("Mapping in a VBO that doesn't exist?");

    memcpy(data.data(), vram, (byte_count / 4) * sizeof(GLfloat));

    glUnmapBuffer(GL_ARRAY_BUFFER);
    glBindBuffer(GL_ARRAY_BUFFER, current_buffer);

    if (!vertex_array_enabled)
        glDisableClientState(GL_VERTEX_ARRAY);
    return data;
}

std::vector<GLuint> JGL::VRamList::GetDataUI() const {

    if (!element_array_buffer)
        JGL::Logger::Warning("Getting data as GLuint but the buffer data is GLfloat?");

    GLint current_buffer = 0;
    glGetIntegerv(GL_ELEMENT_ARRAY_BUFFER_BINDING, &current_buffer);

    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, list_handle);
    std::vector<GLuint> data(byte_count / 4);

    void* vram = glMapBuffer(GL_ELEMENT_ARRAY_BUFFER, GL_READ_ONLY);
    if (vram == nullptr)
        JGL::Logger::Fatal("Mapping in a VBO that doesn't exist?");

    memcpy(data.data(), vram, (byte_count / 4) * sizeof(GLuint));

    glUnmapBuffer(GL_ELEMENT_ARRAY_BUFFER);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, current_buffer);

    return data;
}

JGL::VRamList::VRamList(const GLfloat* data, const long& length, VRamUsageHint hint) {
    usage_hint = hint;
    load(data, (long) sizeof(GLfloat) * length);
}

JGL::VRamList::VRamList(const GLuint* data, const long& length, VRamUsageHint hint) {
    usage_hint = hint;
    load(data, (long) sizeof(GLuint) * length);
}

JGL::VRamList::VRamList(const Vector2* data, const long& length, VRamUsageHint hint) {
    usage_hint = hint;
    load(reinterpret_cast<const GLfloat*>(data), (long) sizeof(Vector2) * length);
}

JGL::VRamList::VRamList(const Vector3* data, const long& length, VRamUsageHint hint) {
    usage_hint = hint;
    load(reinterpret_cast<const GLfloat*>(data), (long) sizeof(Vector3) * length);
}

JGL::VRamList::VRamList(const Vector4* data, const long& length, VRamUsageHint hint) {
    usage_hint = hint;
    load(reinterpret_cast<const GLfloat*>(data), (long) sizeof(Vector4) * length);
}

JGL::VRamList::VRamList(const Color4* data, const long& count, VRamUsageHint hint) {
    usage_hint = hint;
    load(reinterpret_cast<const GLfloat *>(data), (long) sizeof(GLuint) * count);
}
void JGL::VRamList::SetData(const GLfloat* data, const long& length) {
    SetData((void*) data, length);
}

void JGL::VRamList::SetData(const Vector2* data, const long& length) {
    SetData((void*) data, length * 2);
}

void JGL::VRamList::SetData(const Vector3* data, const long& length) {
    SetData((void*) data, length * 3);
}

void JGL::VRamList::SetData(const Vector4* data, const long& length) {
    SetData((void*) data, length * 4);
}

void JGL::VRamList::SetData(const GLuint* data, const long& length) {
    SetData((void*) data, length);
}

void JGL::VRamList::SetData(const Vector2i* data, const long& length) {
    SetData((void*) data, length * 2);
}

void JGL::VRamList::SetData(const Color4* data, const long& count) {
    SetData((void*) data, count);
}

void JGL::VRamList::UpdateData(const GLfloat* data, const long& offset, const long& length) {
    UpdateData((void*) data, offset, length);
}

void JGL::VRamList::UpdateData(const Vector2* data, const long& offset, const long& length) {
    UpdateData((void*) data, offset, length * 2);
}

void JGL::VRamList::UpdateData(const Vector3* data, const long& offset, const long& length) {
    UpdateData((void*) data, offset, length * 3);
}

void JGL::VRamList::UpdateData(const Vector4* data, const long& offset, const long& length) {
    UpdateData((void*) data, offset, length * 4);
}

void JGL::VRamList::UpdateData(const GLuint* data, const long& offset, const long& length) {
    UpdateData((void*) data, offset, length);
}

void JGL::VRamList::UpdateData(const Vector2i* data, const long& offset, const long& length) {
    UpdateData((void*) data, offset, length * 2);
}

void JGL::VRamList::UpdateData(const Color4* data, const long &offset, const long &count) {
    UpdateData((void*) data, offset, count);
}

JGL::VRamList::~VRamList() {
    Erase();
}

JGL::VRamList::VRamList(const JGL::VRamList& rhs) {
    if (rhs.element_array_buffer) {
        auto data_array = rhs.GetDataUI();
        this->load(data_array.data(), data_array.size());
    }

    auto data_array = rhs.GetDataF();
    this->load(data_array.data(), data_array.size());
}

JGL::VRamList::VRamList(const size_t& byte_count, bool element_array_buffer, VRamUsageHint hint) {
    this->element_array_buffer = element_array_buffer;
    this->byte_count = byte_count;

    GLint current_buffer = 0;
    GLenum buffer = element_array_buffer ? GL_ELEMENT_ARRAY_BUFFER : GL_ARRAY_BUFFER;
    GLenum buffer_binding = element_array_buffer ? GL_ELEMENT_ARRAY_BUFFER_BINDING : GL_ARRAY_BUFFER_BINDING;

    glGetIntegerv(buffer_binding, &current_buffer);
    glGenBuffers(1, &list_handle);
    glBindBuffer(buffer, list_handle);
    glBufferData(buffer, byte_count, nullptr, hint);

    glBindBuffer(buffer, current_buffer);
}

void JGL::VRamList::UpdateData(const uint8_t* data, const long& offset, const long& count) {
    GLint current_buffer = 0;
    GLenum buffer_type = element_array_buffer ? GL_ELEMENT_ARRAY_BUFFER : GL_ARRAY_BUFFER;
    GLenum buffer_binding = element_array_buffer ? GL_ELEMENT_ARRAY_BUFFER_BINDING : GL_ARRAY_BUFFER_BINDING;

    glGetIntegerv(buffer_binding, &current_buffer);
    glBindBuffer(buffer_type, list_handle);

    glBufferSubData(buffer_type, offset, count, data);

    glBindBuffer(buffer_type, current_buffer);
}
